''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' Battery Management Program for Caravan 12V System        '
' CVANBMON.BAS                                             '
' Copyright Doug Pankhurst May 2012                        '
' additional routines copyright Geoff Graham 2012          '
' Written for MMBasic V4.2, copyright Geoff Graham 2012    '
''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' File Version:    1.00   10 Jan 2013
'
' This program monitors a 12V Battery System, monitoring the battery voltage,
' the amount of charge supplied into the system via an input current shunt, the
' amount of charge used by the system via a load current shunt and calculates
' the charge/discharge state of the battery by comparing current in versus
' current out. It also estimates time to 30% capacity based on battery voltage
' and current discharge rate and provides a warning.
' Real time data is displayed on VGA display as well as historical
' data recorded to a log file on the SD card for later analysis.


'Disable interrupts and initialise
EntryPoint:  SetTick 0,0
TRUE = 1
FALSE = 0
HIGH = 1
LOW = 0
BVolts = 12.8
Dim BVArray(29)
BatCharge = 100
AverageBV=12.8
Dim BVAArray(59)
DeltaBV = 0.1
Time2Discharge = 20
SVolts = 20
Dim SVArray(29)
AverageSV = 21
Dim SVAArray(59)
InputCurrent = 18
Dim ICArray(29)
AverageIC = 18
Dim ICAArray(59)
LoadCurrent = 12
Dim LCArray(29)
AverageLC = 12
Dim LCAArray(59)
BatCurrent = 6
AverageBC = 6
InputCurrentAH = 0
PreviousICAH = 0
LoadCurrentAH = 0
PreviousLCAH = 0
BatCurrentAH = 0
PreviousBCAH = 5
DayP2BCAH = -2
DayP3BCAH = 3
DayP4BCAH = -5
LogArray$ = ""
NewDataFlag = 0
MinFlag = 0
HourFlag = 0
EODFlag = 0

Entry:
Mode 3,7
Colour 7,0
Cls
Option USB Off
DisplayStatic
CreateLogFile
PreviousICAH = 50
PreviousLCAH = 45
PreviousBCAH = 5
OldABV = 13
LFAPointer = 0
For Ax=0 To 29
BVArray(Ax) = 13
Next Ax
MinTurnover=1
Do While MinTurnover <> 0
MinTurnover=Val(Mid$(Time$,7,2))
Loop
SetTick 2000,GetData   'enable interrupts

Main:
Do While NewDataFlag = 0
Loop

' Two second processing
Print @(1,14)Time$ + "   " + Date$
For Ax = 29 To 1 Step -1
BVArray(Ax) = BVArray(Ax-1)
Next Ax
BVArray(0) = BVolts
For Ax = 29 To 1 Step -1
SVArray(Ax) = SVArray(Ax-1)
Next Ax
SVArray(0) = SVolts
For Ax = 29 To 1 Step -1
ICArray(Ax) = ICArray(Ax-1)
Next Ax
ICArray(0) = InputCurrent
For Ax = 29 To 1 Step -1
LCArray(Ax) = LCArray(Ax-1)
Next Ax
LCArray(0) = LoadCurrent
BatCurrent = InputCurrent-LoadCurrent
InputCurrentAH = InputCurrentAH + (InputCurrent/1800)
LoadCurrentAH = LoadCurrentAH + (LoadCurrent/1800)
BatCurrentAH = InputCurrentAH-LoadCurrentAH
Print @(394,25)Format$(BatCharge,"%3.0f")+"%"
Print @(111,45)Format$(BVolts,"%6.2fV")
Print @(111,56)Format$(SVolts,"%6.2fV")
Print @(111,67)Format$(InputCurrent,"%6.1fA")
Print @(111,78)Format$(LoadCurrent,"%6.1fA")
Print @(111,89)Format$(BatCurrent,"%+6.1fA")
Print @(312,45)Format$(AverageBV,"%6.2fV")
Print @(312,56)Format$(AverageSV,"%6.2fV")
Print @(376,149)Format$(Time2Discharge,"%4.0f min")
Print @(312,67)Format$(AverageIC,"%5.1fA")
Print @(312,78)Format$(AverageLC,"%5.1fA")
Print @(312,89)Format$(AverageBC,"%+5.1fA")
Print @(111,127)Format$(InputCurrentAH,"%6.1fahrs")
Print @(111,138)Format$(LoadCurrentAH,"%6.1fahrs")
Print @(111,149)Format$(BatCurrentAH,"%+6.1fahrs")
Print @(312,127)Format$(PreviousICAH,"%5.1fahrs")
Print @(312,138)Format$(PreviousLCAH,"%5.1fahrs")
Print @(312,149)Format$(PreviousBCAH,"%+5.1fahrs")
Print @(360,67)Format$(DayP2BCAH,"%+5.1fahrs")
Print @(360,78)Format$(DayP3BCAH,"%+5.1fahrs")
Print @(360,89)Format$(DayP4BCAH,"%+5.1fahrs")
NewDataFlag = 0

' Minute processing
If MinFlag Then
OldABV = AverageBV
AverageBV = 0
For Ax = 0 To 29
AverageBV = AverageBV + BVArray(Ax)
Next Ax
AverageBV = AverageBV/30
OldASV = AverageSV
AverageSV = 0
For Ax = 0 To 29
AverageSV = AverageSV + SVArray(Ax)
Next Ax
AverageSV = AverageSV/30
AverageIC = 0
For Ax = 0 To 29
AverageIC = AverageIC + ICArray(Ax)
Next Ax
AverageIC = AverageIC/30
AverageLC = 0
For Ax = 0 To 29
AverageLC = AverageLC + LCArray(Ax)
Next Ax
AverageLC = AverageLC/30
AverageBC = AverageIC - AverageLC
BatCharge = Int((AverageBV - 11.75) * 100)
If BatCharge < 30 Then BatCharge = 30
If BatCharge > 100 Then BatCharge = 100
FillBarGraph(BatCharge)
DeltaBV = OldABV - AverageBV
If Sgn(DeltaBV) = +1 Then
DVRate = BVArray(0)
Time2Discharge = 0
Do While DVRate > 12.1
DVRate = DVRate - DeltaBV
Time2Discharge = Time2Discharge + 1
Loop
Else
Print @(376,149)"Charging "
EndIf
BVAArray(LFAPointer) = AverageBV
SVAArray(LFAPointer) = AverageSV
ICAArray(LFAPointer) = AverageIC
LCAArray(LFAPointer) = AverageLC
LFAPointer = LFAPointer + 1
rg_mins = (((Val(Mid$(Time$,1,2)) Mod 6)*60))+(Val(Mid$(Time$,4,2)))
rg_mins = rg_mins + 30
rg_abv = 413 - Int((AverageBV-7.5) * 20)
if rg_abv > 413 then rg_abv = 413
rg_asv = 413 - Int((AverageSV-7.5) * 20)
if rg_asv > 413 then rg_asv = 413
if rg_asv < 163 then rg_asv = 163
rg_aic = 413 - Int(AverageIC * 5)
rg_alc = 413 - Int(AverageLC * 5)
If (Sgn(AverageBC) = +1) Then
rg_abc = 413 - Int(AverageBC * 5)
Else
rg_abc = 413 + Int(AverageBC * 5)
EndIf
For rg_y = 163 To 413 Step 1
If rg_abc = rg_y Then
If (Sgn(AverageBC) = +1) Then
Pixel(rg_mins,rg_y) = 2
Else
Pixel(rg_mins,rg_y) = 4
EndIf
Else
Pixel(rg_mins,rg_y) = 0
EndIf
If rg_abv = rg_y Then
Pixel(rg_mins,rg_y) = 7
EndIf
If rg_asv = rg_y Then
Pixel(rg_mins,rg_y) = 1
EndIf
If rg_aic = rg_y Then
Pixel(rg_mins,rg_y) = 6
EndIf
If rg_alc = rg_y Then
Pixel(rg_mins,rg_y) = 5
EndIf
Next rg_y
For rg_y = 163 To 413 Step 1
Pixel(rg_mins+1,rg_y) = 0
Pixel(rg_mins+2,rg_y) = 0
Pixel(rg_mins+3,rg_y) = 0
Next rg_y
MinFlag = 0
EndIf

' Hour processing
If HourFlag Then
LogArray$ = ""
For x=0 To 59 Step 1
LogArray$ = ""
LogArray$ = LogArray$+Format$(BVAArray(x),"%+6.2fV")+Chr$(44)
LogArray$ = LogArray$+Format$(SVAArray(x),"%+6.2fV")+Chr$(44)
LogArray$ = LogArray$+Format$(ICAArray(x),"%+5.1fA")+Chr$(44)
LogArray$ = LogArray$+Format$(LCAArray(x),"%+5.1fA")
Print #1,LogArray$
Next x
LFAPointer = 0
HourFlag = 0
EndIf

' End of Day processing
If EODFlag Then
Close #1
PreviousICAH = InputCurrentAH
InputCurrentAH = 0
PreviousLCAH = LoadCurrentAH
LoadCurrentAH = 0
DayP4BCAH = DayP3BCAH
DayP3BCAH = DayP2BCAH
DayP2BCAH = PreviousBCAH
PreviousBCAH = BatCurrentAH
BatCurrentAH = 0
CreateLogFile
EODFlag = 0
EndIf
GoTo Main

' Interrupt Routine
GetData:
SetPin 1,1
SetPin 2,1
SetPin 4,1
SetPin 3,1
ic=0
BVolts = 0
GetPin4 = 0
For ic = 1 To 10
GetPin4 = Pin(4)
BVolts = BVolts + GetPin4
Next ic
BVolts = BVolts/10
BVolts = BVolts*6
ic=0
SVolts=0
GetPin3 = 0
For ic = 1 To 10
GetPin3 = Pin(3)
SVolts = SVolts + GetPin3
Next ic
SVolts = SVolts/10
SVolts = SVolts*10
ic=0
InputCurrent = 0
GetPin1 = 0
For ic = 1 To 10
GetPin1 = Pin(1)
InputCurrent = InputCurrent + GetPin1
Next ic
InputCurrent = InputCurrent/10
InputCurrent = InputCurrent*20
ic=0
LoadCurrent = 0
GetPin2 = 0
For ic = 1 To 10
GetPin2 = Pin(2)
LoadCurrent = LoadCurrent + GetPin2
Next ic
LoadCurrent = LoadCurrent/10
LoadCurrent = LoadCurrent*20
If Mid$(Time$,7,2) = "00" Then
MinFlag = 1
EndIf
CurrentMins$ = Mid$(Time$,4,2)
If Mid$(Time$,7,2) = "00" And Mid$(Time$,4,2) = "00" Then
HourFlag = 1
EndIf
CurrentHrs$ = Mid$(Time$,1,2)
If Mid$(Time$,7,2)="00" And Mid$(Time$,4,2)="00" And Mid$(Time$,1,2)="00" Then
EODFlag = 1
EndIf
NewDataFlag = 1
IReturn
' end of interrupt routine

' Subroutines
Sub PrintStatus Status$
Print @(348,14)"            "
Print @(348,14)Status$
End Sub


Sub DisplayStatic
Cls
Print @(150,1)"CARAVAN POWER MONITOR V1.00"
Print @(1,25)"Instantaneous Readings"
Print @(198,25)"Minute Average Readings"
Line(1,36)-(158,36)
Line(198,36)-(350,36)
Print @(384,1)"Battery"
Print @(1,45)"Battery Volts     ="
Print @(198,45)"Battery Volts     ="
Print @(387,12)"Charge"
Print @(1,56)"Solar Array Volts ="
Print @(198,56)"Solar Array Volts ="
Print @(1,67)"Input Current     ="
Print @(198,67)"Input Current     ="
Print @(1,78)"Load Current      ="
Print @(198,78)"Load Current      ="
Print @(1,89)"Bat Chg/Dis Curr  ="
Print @(198,89)"Bat Chg/Dis Curr  ="
Print @(1,108)"Daily Accumulating Readings"
Print @(198,108)"Previous Days Total Readings"
Print @(369,45)"Bat Charge"
Print @(366,56)"Last 3 Days"
Line(1,120)-(170,120)
Line(198,120)-(362,120)
Print @(1,127)"Input Charge      ="
Print @(198,127)"Input Charge      ="
Print @(384,127)"Time to"
Print @(1,138)"Load Discharge    ="
Print @(198,138)"Load Discharge    ="
Print @(400,138)"30%"
Print @(1,149)"Battery Chg/DisChg="
Print @(198,149)"Battery Chg/DisChg="
Line(0,162)-(479,431),1,B
Colour 7,0
Print @(424,164)"Battery"
Print @(430,174)"Volts"
Colour 1,0
Print @(430,194)"Solar"
Print @(430,204)"Volts"
Colour 6,0
Print @(430,224)"Input"
Print @(433,234)"Amps"
Colour 5,0
Print @(433,254)"Load"
Print @(433,264)"Amps"
Colour 4,0
Print @(427,284)"Battery"
Print @(424,294)"Amps Out"
Colour 2,0
Print @(427,314)"Battery"
Print @(427,324)"Amps In"
Colour 7,0
volts=20
amps = 50
For ry =163 to 403 step 10
Print @(400,ry)Format$(amps,"%2.0fA")
amps=amps-2
next ry
For ry =163 to 403 step 20
Print @(2,ry)Format$(volts,"%2.0fV")
volts=volts-1
next ry
Line(29,163)-(29,413)
Line(391,163)-(391,413)
For yrg = 163 To 413 Step 10
Pixel(28,yrg)=7
Pixel(27,yrg)=7
Pixel(392,yrg)=7
Pixel(393,yrg)=7
Next yrg
For yrg = 163 To 413 Step 20
Pixel(27,yrg)=7
Next yrg
Line(30,414)-(390,414)
For xrg = 30 To 390 Step 5
Pixel(xrg,415)=7
Next xrg
For xrg = 30 To 390 Step 15
Pixel(xrg,415)=7
Pixel(xrg,416)=7
Next xrg
For xrg = 30 To 390 Step 30
Pixel(xrg,415)=7
Pixel(xrg,416)=7
Pixel(xrg,417)=7
Next xrg
For xrg = 30 To 390 Step 60
Pixel(xrg,415)=7
Pixel(xrg,416)=7
Pixel(xrg,417)=7
Pixel(xrg,418)=7
Next xrg
x_label=1
for bx = 88 to 388 step 60
Print @(bx,418)chr$(x_label)
x_label=x_label+1
next bx
Local xtl,ytl,xbr,ybr,percent,xpos,ypos,y
xtl=432:ytl=0:xbr=xtl+47:ybr=ytl+162
Line(xtl,ytl)-(xbr,ybr),1,B
percent=100
Print @(xtl+3,ytl+2)"Battery"
Print @(xtl+13,ytl+13)100
Print @(xtl+40,ytl+13)"%"
percent=percent-5
For y=ytl+22 To 142 Step 10
Print @(xtl+19,y)percent
Print @(xtl+40,y)"%"
percent = percent-5
Next y
End Sub


Sub CreateLogFile
LogFileName$="B:"+Mid$(Date$,9,2)+Mid$(Date$,4,2)+Left$(Date$,2)+".CSV"
Open LogFileName$ For Output As #1
Print #1, "Battery Voltage";Chr$(44);
Print #1, "Solar Array Voltage";Chr$(44);
Print #1, "Input Current";Chr$(44);
Print #1, "Load Current"
End Sub


Sub FillBarGraph(BatCharge)
Local xtl,ytl,bat_chg,xpos,ypos,yper,xright,xleft,ybot,ytop
bat_chg=BatCharge
xtl=432:ytl=0:xbr=xtl+47:ybr=ytl+162
If bat_chg < 31 Then
bat_chg=30
Colour 4,7
Print @(xtl + 4,ytl + 151,2)"DANGER"
Else
Colour 2,7
Print @(xtl + 4,ytl + 151,2)"  OK  "
EndIf
Colour 7,0
If bat_chg > 100 Then bat_chg=100
ytop=12
yper=Int(((100 - bat_chg) * 1.97) + 12)
ybot=150
xleft=437
xright=446
For ypos=ybot To ytop Step -1
If yper < ypos Then
For xpos=xleft To xright Step 1
Pixel(xpos,ypos)=2
Next xpos
Else
For xpos=xleft To xright Step 1
Pixel(xpos,ypos)=0
Next xpos
EndIf
Next ypos
End Sub
